/*
 *  Copyright (C) 2019 KeePassXC Team <team@keepassxc.org>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 or (at your option)
 *  version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Help.h"

#include "Utils.h"

Help::Help()
{
    name = QString("help");
    description = QObject::tr("Display command help.");
}

int Help::execute(const QStringList& arguments)
{
    auto& out = Utils::STDOUT;
    QSharedPointer<Command> command;
    if (arguments.size() > 1 && (command = Commands::getCommand(arguments.at(1)))) {
        out << command->getHelpText();
    } else {
        out << "\n\n" << QObject::tr("Available commands:") << "\n";
        for (auto& cmd : Commands::getCommands()) {
            out << cmd->getDescriptionLine();
        }
    }
    out.flush();
    return EXIT_SUCCESS;
}
