/*
 *  Copyright (C) 2020 Aetf <aetf@unlimitedcodeworks.xyz>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 or (at your option)
 *  version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "FdoSecretsProxy.h"

#define IMPL_PROXY(name)                                                                                               \
    name##Proxy::name##Proxy(                                                                                          \
        const QString& service, const QString& path, const QDBusConnection& connection, QObject* parent)               \
        : QDBusAbstractInterface(service, path, staticInterfaceName(), connection, parent)                             \
    {                                                                                                                  \
    }                                                                                                                  \
    name##Proxy::~name##Proxy() = default;

IMPL_PROXY(Service)
IMPL_PROXY(Collection)
IMPL_PROXY(Item)
IMPL_PROXY(Session)
IMPL_PROXY(Prompt)

#undef IMPL_PROXY
